import os
import sys
import threading
import traceback


from flask import g, request

import odio

from sqlalchemy import select
from sqlalchemy.orm import joinedload

from werkzeug.exceptions import BadRequest

import chellow.dloads
from chellow.models import MeasurementRequirement, Session, Ssc, Tpr, User
from chellow.utils import req_bool
from chellow.views import chellow_redirect


def write_spreadsheet(
    fl,
    compressed,
    ssc_rows,
    mr_rows,
    tpr_rows,
):
    fl.seek(0)
    fl.truncate()
    with odio.create_spreadsheet(fl, "1.2", compressed=compressed) as f:
        f.append_table("SSCs", ssc_rows)
        f.append_table("MRs", mr_rows)
        f.append_table("TPRs", tpr_rows)


def content(
    user_id,
    compression,
):
    sess = rf = None
    ssc_rows = [
        [
            "ssc_code",
            "ssc_description",
            "is_import",
            "valid_from",
            "valid_to",
        ]
    ]
    mr_rows = [["ssc_code", "tpr_code"]]
    tpr_rows = [["tpr_code", "is_teleswitch", "is_gmt"]]
    try:
        with Session() as sess:
            user = User.get_by_id(sess, user_id)

            running_name, finished_name = chellow.dloads.make_names("sscs.ods", user)

            rf = open(running_name, "wb")

            for ssc in sess.scalars(select(Ssc).order_by(Ssc.code)):
                ssc_rows.append(
                    (
                        ssc.code,
                        ssc.description,
                        ssc.is_import,
                        ssc.valid_from,
                        ssc.valid_to,
                    )
                )
            for mr in sess.scalars(
                select(MeasurementRequirement)
                .join(Ssc)
                .join(Tpr)
                .order_by(Ssc.code, Tpr.code)
                .options(
                    joinedload(MeasurementRequirement.ssc),
                    joinedload(MeasurementRequirement.tpr),
                )
            ):
                mr_rows.append((mr.ssc.code, mr.tpr.code))
            for tpr in sess.scalars(select(Tpr).order_by(Tpr.code)):
                tpr_rows.append((tpr.code, tpr.is_teleswitch, tpr.is_gmt))

            write_spreadsheet(
                rf,
                compression,
                ssc_rows,
                mr_rows,
                tpr_rows,
            )
    except BadRequest as e:
        msg = e.description + traceback.format_exc()
        sys.stderr.write(msg + "\n")
        ssc_rows.append(["Problem " + msg])
        write_spreadsheet(rf, compression, ssc_rows, mr_rows, tpr_rows)
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg + "\n")
        ssc_rows.append(["Problem " + msg])
        if rf is None:
            msg = traceback.format_exc()
            r_name, f_name = chellow.dloads.make_names("error.txt", None)
            ef = open(r_name, "w")
            ef.write(msg + "\n")
            ef.close()
        else:
            write_spreadsheet(rf, compression, ssc_rows, mr_rows, tpr_rows)
    finally:
        if rf is not None:
            rf.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    compression = req_bool("compression") if "compression" in request.values else True

    user = g.user

    args = (user.id, compression)
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



import chellow.e.computer
import chellow.reports.report_181
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_write_sites(mocker):
    sess = mocker.Mock()
    caches = {}
    writer = mocker.Mock()
    year = 2010
    year_start = to_utc(ct_datetime(year - 1, 4, 1))
    year_finish = to_utc(ct_datetime(year, 3, 31, 23, 30))
    month_start = to_utc(utc_datetime(year, 3, 1))

    site_id = None
    site = mocker.Mock()
    era = mocker.Mock()
    source_codes = ("gen", "gen-net")

    ms = mocker.patch("chellow.reports.report_181._make_sites", autospec=True)
    ms.return_value = [site]
    de = mocker.patch("chellow.e.computer.displaced_era", autospec=True)
    de.return_value = era
    ss = mocker.patch("chellow.e.computer.SiteSource", autospec=True)
    ss_instance = ss.return_value
    ss_instance.hh_data = [
        {
            "start-date": month_start,
            "triad-actual-1-date": utc_datetime(year, 1, 12),
            "triad-actual-1-msp-kw": 12,
            "triad-actual-1-laf": 1,
            "triad-actual-1-gsp-kw": 12,
            "triad-actual-2-date": utc_datetime(year, 1, 12),
            "triad-actual-2-msp-kw": 12,
            "triad-actual-2-laf": 1,
            "triad-actual-2-gsp-kw": 12,
            "triad-actual-3-date": utc_datetime(year, 1, 12),
            "triad-actual-3-msp-kw": 12,
            "triad-actual-3-laf": 1,
            "triad-actual-3-gsp-kw": 12,
            "triad-actual-gsp-kw": 12,
            "triad-actual-rate": 10,
            "triad-actual-gbp": 120,
        }
    ]

    ss_instance.supplier_bill_hhs = {month_start: {}}

    mocker.patch("chellow.e.duos.duos_vb", autospec=True)
    mocker.patch("chellow.e.tnuos.hh", autospec=True)

    forecast_date = chellow.e.computer.forecast_date()

    chellow.reports.report_181._write_sites(sess, caches, writer, year, site_id)

    ms.assert_called_once_with(sess, year_start, year_finish, site_id, source_codes)
    de.assert_called_once_with(
        sess, caches, site, month_start, year_finish, forecast_date
    )



from io import StringIO

from chellow.models import (
    Comm,
    Contract,
    Cop,
    EnergisationStatus,
    GspGroup,
    MarketRole,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfc,
    MtcParticipant,
    Participant,
    Pc,
    Site,
    Source,
    User,
    UserRole,
    VoltageLevel,
    insert_comms,
    insert_cops,
    insert_energisation_statuses,
    insert_sources,
    insert_voltage_levels,
)
from chellow.reports.report_29 import content
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_content(mocker, sess):
    vf = to_utc(ct_datetime(2000, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(sess, "Mop", participant, "", {}, vf, None, {})
    dc_contract = Contract.insert_dc(sess, "DC 2000", participant, "", {}, vf, None, {})
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, vf, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        utc_datetime(1996, 1, 1),
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(sess, "510", "HV", voltage_level, False, True, vf, None)
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "net")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        "773",
        dc_contract,
        "ghyy3",
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        {},
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    era = supply.eras[0]
    channel = era.insert_channel(sess, True, "ACTIVE")
    data_raw = [
        {
            "start_date": to_utc(ct_datetime(2010, 1, 1)),
            "value": 10,
            "status": "A",
        }
    ]
    channel.add_hh_data(sess, data_raw)
    editor = UserRole.insert(sess, "editor")
    user = User.insert(sess, "admin@example.com", "xxx", editor, None)

    sess.commit()
    f = StringIO()
    f.close = mocker.Mock()
    mocker.patch("chellow.reports.report_29.open", return_value=f)
    mocker.patch("chellow.reports.report_29.os.rename")
    mocker.patch(
        "chellow.reports.report_29.chellow.dloads.make_names", return_value=("a", "b")
    )
    start_date = to_utc(ct_datetime(2010, 1, 1))
    finish_date = to_utc(ct_datetime(2010, 1, 31, 23, 30))
    site_id = site.id
    typ = "used"
    content(start_date, finish_date, site_id, typ, user.id)
    expected = [
        [
            "site_code",
            "type",
            "hh_start_clock_time",
            "total",
            "1",
            "2",
            "3",
            "4",
            "5",
            "6",
            "7",
            "8",
            "9",
            "10",
            "11",
            "12",
            "13",
            "14",
            "15",
            "16",
            "17",
            "18",
            "19",
            "20",
            "21",
            "22",
            "23",
            "24",
            "25",
            "26",
            "27",
            "28",
            "29",
            "30",
            "31",
            "32",
            "33",
            "34",
            "35",
            "36",
            "37",
            "38",
            "39",
            "40",
            "41",
            "42",
            "43",
            "44",
            "45",
            "46",
            "47",
            "48",
            "49",
            "50",
        ],
        [
            "CI017",
            "used",
            "2010-01-01",
            "10",
            "10",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "0",
            "",
            "",
        ],
    ]
    assert f.getvalue().startswith("".join(",".join(line) + "\n" for line in expected))



import csv
import os
import threading
import traceback

from flask import g, request

from sqlalchemy import and_, or_
from sqlalchemy.orm import joinedload

from werkzeug.exceptions import BadRequest

import chellow.dloads
from chellow.models import Batch, Bill, BillType, Era, RegisterRead, Session, Supply
from chellow.utils import c_months_u, csv_make_val, req_int
from chellow.views import chellow_redirect


def content(year, month, months, supply_id, user):
    f = None
    try:
        with Session() as sess:
            running_name, finished_name = chellow.dloads.make_names(
                "register_reads.csv", user
            )
            f = open(running_name, mode="w", newline="")
            w = csv.writer(f, lineterminator="\n")
            titles = (
                "Duration Start",
                "Duration Finish",
                "Supply Id",
                "Import MPAN Core",
                "Export MPAN Core",
                "Batch Reference",
                "Bill Id",
                "Bill Reference",
                "Bill Issue Date",
                "Bill Type",
                "Register Read Id",
                "TPR",
                "Coefficient",
                "Previous Read Date",
                "Previous Read Value",
                "Previous Read Type",
                "Present Read Date",
                "Present Read Value",
                "Present Read Type",
            )
            w.writerow(titles)

            month_pairs = list(
                c_months_u(finish_year=year, finish_month=month, months=months)
            )
            start_date, finish_date = month_pairs[0][0], month_pairs[-1][-1]

            supplies = (
                sess.query(Supply)
                .join(Bill)
                .join(RegisterRead)
                .filter(
                    or_(
                        and_(
                            RegisterRead.present_date >= start_date,
                            RegisterRead.present_date <= finish_date,
                        ),
                        and_(
                            RegisterRead.previous_date >= start_date,
                            RegisterRead.previous_date <= finish_date,
                        ),
                    )
                )
                .order_by(Bill.supply_id)
                .distinct()
            )

            if supply_id is not None:
                supply = Supply.get_by_id(sess, supply_id)
                supplies = supplies.filter(Bill.supply == supply)

            for supply in supplies:
                supply_id = supply.id
                for bill, batch, bill_type in (
                    sess.query(Bill, Batch, BillType)
                    .join(Batch)
                    .join(BillType)
                    .join(RegisterRead)
                    .filter(
                        Bill.supply == supply,
                        or_(
                            and_(
                                RegisterRead.present_date >= start_date,
                                RegisterRead.present_date <= finish_date,
                            ),
                            and_(
                                RegisterRead.previous_date >= start_date,
                                RegisterRead.previous_date <= finish_date,
                            ),
                        ),
                    )
                ):
                    era = supply.find_era_at(sess, bill.start_date)
                    if era is None:
                        eras = (
                            sess.query(Era)
                            .filter(Era.supply == supply)
                            .order_by(Era.start_date)
                            .all()
                        )
                        if bill.start_date < eras[0].start_date:
                            era = eras[0]
                        else:
                            era = eras[-1]

                    for read in (
                        sess.query(RegisterRead)
                        .filter(
                            RegisterRead.bill == bill,
                            or_(
                                and_(
                                    RegisterRead.present_date >= start_date,
                                    RegisterRead.present_date <= finish_date,
                                ),
                                and_(
                                    RegisterRead.previous_date >= start_date,
                                    RegisterRead.previous_date <= finish_date,
                                ),
                            ),
                        )
                        .options(
                            joinedload(RegisterRead.tpr),
                            joinedload(RegisterRead.previous_type),
                            joinedload(RegisterRead.present_type),
                        )
                    ):
                        vals = [
                            start_date,
                            finish_date,
                            supply_id,
                            era.imp_mpan_core,
                            era.exp_mpan_core,
                            batch.reference,
                            bill.id,
                            bill.reference,
                            bill.issue_date,
                            bill_type.code,
                            read.id,
                            "md" if read.tpr is None else read.tpr.code,
                            read.coefficient,
                            read.previous_date,
                            read.previous_value,
                            read.previous_type.code,
                            read.present_date,
                            read.present_value,
                            read.present_type.code,
                        ]
                        w.writerow(csv_make_val(v) for v in vals)

                    # Avoid a long-running transaction
                    sess.rollback()

    except BadRequest as e:
        w.writerow([e.description])
    except BaseException:
        msg = traceback.format_exc()
        f.write(msg)
    finally:
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    year = req_int("end_year")
    month = req_int("end_month")
    months = req_int("months")
    supply_id = req_int("supply_id") if "supply_id" in request.values else None
    args = (year, month, months, supply_id, g.user)
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



import os
import sys
import threading
import traceback

from flask import g, request

import odio

from sqlalchemy import or_, true
from sqlalchemy.orm import joinedload
from sqlalchemy.sql.expression import null

from werkzeug.exceptions import BadRequest

import chellow.dloads
import chellow.e.computer
from chellow.e.computer import contract_func
from chellow.gas.engine import GDataSource
from chellow.models import (
    GBill,
    GContract,
    GEra,
    GSupply,
    Session,
    Site,
    SiteGEra,
)
from chellow.utils import (
    c_months_u,
    ct_datetime_now,
    hh_format,
    hh_max,
    hh_min,
    make_val,
    req_bool,
    req_int,
)
from chellow.views import chellow_redirect


CATEGORY_ORDER = {None: 0, "unmetered": 1, "nhh": 2, "amr": 3, "hh": 4}
meter_order = {"hh": 0, "amr": 1, "nhh": 2, "unmetered": 3}


def write_spreadsheet(fl, compressed, site_rows, era_rows):
    fl.seek(0)
    fl.truncate()
    with odio.create_spreadsheet(fl, "1.2", compressed=compressed) as f:
        f.append_table("Site Level", site_rows)
        f.append_table("Era Level", era_rows)


def content(
    site_id, g_supply_id, user, compression, finish_year, finish_month, months, now=None
):
    if now is None:
        now = ct_datetime_now()
    report_context = {}
    month_list = list(
        c_months_u(finish_year=finish_year, finish_month=finish_month, months=months)
    )
    start_date, finish_date = month_list[0][0], month_list[-1][-1]

    try:
        with Session() as sess:
            base_name = [
                "g_monthly_duration",
                hh_format(start_date)
                .replace(" ", "_")
                .replace(":", "")
                .replace("-", ""),
                "for",
                str(months),
                "months",
            ]

            forecast_from = chellow.e.computer.forecast_date()

            sites = (
                sess.query(Site)
                .join(SiteGEra)
                .join(GEra)
                .filter(SiteGEra.is_physical == true())
                .distinct()
                .order_by(Site.code)
            )
            if site_id is not None:
                site = Site.get_by_id(sess, site_id)
                sites = sites.filter(Site.id == site.id)
                base_name.append("site")
                base_name.append(site.code)
            if g_supply_id is not None:
                g_supply = GSupply.get_by_id(sess, g_supply_id)
                base_name.append("g_supply")
                base_name.append(str(g_supply.id))
                sites = sites.filter(GEra.g_supply == g_supply)

            running_name, finished_name = chellow.dloads.make_names(
                "_".join(base_name) + ".ods", user
            )

            rf = open(running_name, "wb")
            site_rows = []
            g_era_rows = []

            era_header_titles = [
                "creation_date",
                "mprn",
                "supply_name",
                "exit_zone",
                "msn",
                "unit",
                "contract",
                "site_id",
                "site_name",
                "associated_site_ids",
                "month",
            ]
            site_header_titles = [
                "creation_date",
                "site_id",
                "site_name",
                "associated_site_ids",
                "month",
            ]
            summary_titles = ["kwh", "gbp", "billed_kwh", "billed_gbp"]

            vb_titles = []
            conts = (
                sess.query(GContract)
                .join(GEra)
                .join(GSupply)
                .filter(
                    GEra.start_date <= finish_date,
                    or_(GEra.finish_date == null(), GEra.finish_date >= start_date),
                )
                .distinct()
                .order_by(GContract.id)
            )
            if g_supply_id is not None:
                conts = conts.filter(GEra.g_supply_id == g_supply_id)
            for cont in conts:
                title_func = chellow.e.computer.contract_func(
                    report_context, cont, "virtual_bill_titles"
                )
                if title_func is None:
                    raise Exception(
                        f"For the contract {cont.name} there doesn't seem to be "
                        f"a 'virtual_bill_titles' function."
                    )
                for title in title_func():
                    if title not in vb_titles:
                        vb_titles.append(title)

            g_era_rows.append(era_header_titles + summary_titles + vb_titles)
            site_rows.append(site_header_titles + summary_titles)

            for month_start, month_finish in month_list:
                for site in sites.filter(
                    GEra.start_date <= month_finish,
                    or_(GEra.finish_date == null(), GEra.finish_date >= month_start),
                ):
                    site_kwh = site_gbp = site_billed_kwh = site_billed_gbp = 0

                    for g_era in (
                        sess.query(GEra)
                        .join(SiteGEra)
                        .filter(
                            SiteGEra.site == site,
                            SiteGEra.is_physical == true(),
                            GEra.start_date <= month_finish,
                            or_(
                                GEra.finish_date == null(),
                                GEra.finish_date >= month_start,
                            ),
                        )
                        .options(
                            joinedload(GEra.g_contract),
                            joinedload(GEra.g_supply),
                            joinedload(GEra.g_supply).joinedload(GSupply.g_exit_zone),
                        )
                        .order_by(GEra.id)
                    ):
                        g_supply = g_era.g_supply

                        if g_supply_id is not None and g_supply.id != g_supply_id:
                            continue

                        ss_start = hh_max(g_era.start_date, month_start)
                        ss_finish = hh_min(g_era.finish_date, month_finish)

                        ss = GDataSource(
                            sess,
                            ss_start,
                            ss_finish,
                            forecast_from,
                            g_era,
                            report_context,
                            None,
                        )

                        contract = g_era.g_contract
                        vb_function = contract_func(
                            report_context, contract, "virtual_bill"
                        )
                        if vb_function is None:
                            raise BadRequest(
                                "The contract "
                                + contract.name
                                + " doesn't have the virtual_bill() function."
                            )
                        vb_function(ss)
                        bill = ss.bill

                        try:
                            gbp = bill["net_gbp"]
                        except KeyError:
                            gbp = 0
                            bill["problem"] += (
                                "For the supply "
                                + ss.mprn
                                + " the virtual bill "
                                + str(bill)
                                + " from the contract "
                                + contract.name
                                + " does not contain the net_gbp key."
                            )
                        try:
                            kwh = bill["kwh"]
                        except KeyError:
                            kwh = 0
                            bill["problem"] += (
                                "For the supply "
                                + ss.mprn
                                + " the virtual bill "
                                + str(bill)
                                + " from the contract "
                                + contract.name
                                + " does not contain the 'kwh' key."
                            )

                        billed_kwh = billed_gbp = 0

                        g_era_associates = {
                            s.site.code for s in g_era.site_g_eras if not s.is_physical
                        }

                        for g_bill in sess.query(GBill).filter(
                            GBill.g_supply == g_supply,
                            GBill.start_date <= ss_finish,
                            GBill.finish_date >= ss_start,
                        ):
                            bill_start = g_bill.start_date
                            bill_finish = g_bill.finish_date
                            bill_duration = (
                                bill_finish - bill_start
                            ).total_seconds() + (30 * 60)
                            overlap_duration = (
                                min(bill_finish, ss_finish) - max(bill_start, ss_start)
                            ).total_seconds() + (30 * 60)
                            overlap_proportion = overlap_duration / bill_duration
                            billed_kwh += overlap_proportion * float(g_bill.kwh)
                            billed_gbp += overlap_proportion * float(g_bill.net)

                        associated_site_ids = ",".join(sorted(g_era_associates))
                        g_era_rows.append(
                            [
                                make_val(v)
                                for v in [
                                    now,
                                    g_supply.mprn,
                                    g_supply.name,
                                    g_supply.g_exit_zone.code,
                                    g_era.msn,
                                    g_era.g_unit.code,
                                    contract.name,
                                    site.code,
                                    site.name,
                                    associated_site_ids,
                                    month_finish,
                                    kwh,
                                    gbp,
                                    billed_kwh,
                                    billed_gbp,
                                ]
                            ]
                            + [make_val(bill.get(t)) for t in vb_titles]
                        )

                        site_kwh += kwh
                        site_gbp += gbp
                        site_billed_kwh += billed_kwh
                        site_billed_gbp += billed_gbp

                    linked_sites = ", ".join(
                        s.code
                        for s in site.find_linked_sites(sess, month_start, month_finish)
                    )

                    site_rows.append(
                        [
                            make_val(v)
                            for v in [
                                now,
                                site.code,
                                site.name,
                                linked_sites,
                                month_finish,
                                site_kwh,
                                site_gbp,
                                site_billed_kwh,
                                site_billed_gbp,
                            ]
                        ]
                    )
                    sess.rollback()
                write_spreadsheet(rf, compression, site_rows, g_era_rows)

    except BadRequest as e:
        site_rows.append(["Problem " + e.description])
        write_spreadsheet(rf, compression, site_rows, g_era_rows)
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg + "\n")
        site_rows.append(["Problem " + msg])
        write_spreadsheet(rf, compression, site_rows, g_era_rows)
    finally:
        try:
            rf.close()
            os.rename(running_name, finished_name)
        except BaseException:
            msg = traceback.format_exc()
            r_name, f_name = chellow.dloads.make_names("error.txt", user)
            ef = open(r_name, "w")
            ef.write(msg + "\n")
            ef.close()


def do_get(sess):
    finish_year = req_int("finish_year")
    finish_month = req_int("finish_month")
    months = req_int("months")

    site_id = req_int("site_id") if "site_id" in request.values else None

    if "g_supply_id" in request.values:
        g_supply_id = req_int("g_supply_id")
    else:
        g_supply_id = None

    if "compression" in request.values:
        compression = req_bool("compression")
    else:
        compression = True

    user = g.user
    args = (site_id, g_supply_id, user, compression, finish_year, finish_month, months)

    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



import typing as t
from functools import lru_cache

from chellow.models import Contract, Session


if t.TYPE_CHECKING:
    from wsgiref.types import StartResponse
    from wsgiref.types import WSGIApplication
    from wsgiref.types import WSGIEnvironment


@lru_cache()
def _get_proto_override():
    with Session() as sess:
        config_contract = Contract.get_non_core_by_name(sess, "configuration")
        props = config_contract.make_properties()
        proto = props.get("redirect_scheme")
    return proto


class MsProxy:
    def __init__(
        self,
        app: "WSGIApplication",
    ) -> None:
        self.app = app
        self.x_proto = None

    def __call__(
        self, environ: "WSGIEnvironment", start_response: "StartResponse"
    ) -> t.Iterable[bytes]:
        try:
            x_proto = environ["HTTP_X_FORWARDED_PROTO"]
        except KeyError:
            x_proto = _get_proto_override()

        if x_proto is not None:
            environ["wsgi.url_scheme"] = x_proto

        x_host = environ.get("HTTP_X_FORWARDED_HOST")
        if x_host is not None:
            environ["HTTP_HOST"] = x_host
            parts = x_host.split(":", 1)
            environ["SERVER_NAME"] = parts[0]
            if len(parts) == 2:
                environ["SERVER_PORT"] = parts[1]

        return self.app(environ, start_response)



from sqlalchemy import select

from chellow.e.mdd_importer import (
    _import_MTC_in_PES_Area,
    _import_Market_Participant_Role,
    _import_Measurement_Requirement,
    _import_Meter_Timeswitch_Class,
    _import_Time_Pattern_Regime,
    _import_Valid_MTC_LLFC_Combination,
    _import_Valid_MTC_LLFC_SSC_Combination,
    _import_Valid_MTC_LLFC_SSC_PC_Combination,
    _import_Valid_MTC_SSC_Combination,
    rate_server_import,
)
from chellow.models import (
    Contract,
    MarketRole,
    MeasurementRequirement,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfcSsc,
    MtcParticipant,
    MtcSsc,
    Participant,
    Pc,
    Ssc,
    Tpr,
    VoltageLevel,
    insert_voltage_levels,
)
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_import_Market_Participant_Role_R(sess):
    participant_code = "EDFI"
    Participant.insert(sess, participant_code, "hyde participant")
    market_role_code = "R"
    MarketRole.insert(sess, market_role_code, "dno")
    sess.commit()

    rows = [
        [
            participant_code,
            market_role_code,
            "16/04/2009",
            "17/12/2014",
            "UK Power Networks (IDNO)",
            "Energy House",
            "Hazelwick Avenue",
            "Three Bridges",
            "Crawley",
            "West Sussex",
            "",
            "",
            "",
            "RH10 1EX",
            "28",
        ]
    ]
    ctx = {}
    _import_Market_Participant_Role(sess, rows, ctx)


def test_import_Measurement_Requirement(sess):
    Ssc.insert(sess, "0001", "All", True, to_utc(ct_datetime(2000, 1, 1)), None)
    Tpr.insert(sess, "00205", True, True)
    rows = [["0001", "00205"]]
    ctx = {}
    _import_Measurement_Requirement(sess, rows, ctx)


def test_import_Measurement_Requirement_no_ongoing_ssc(sess):
    """Check that we get the latest SSC, even when there isn't an ongoing one"""
    Ssc.insert(
        sess,
        "0001",
        "All",
        True,
        to_utc(ct_datetime(2000, 1, 1)),
        to_utc(ct_datetime(2001, 1, 1, 23, 30)),
    )
    latest_ssc = Ssc.insert(
        sess,
        "0001",
        "All",
        True,
        to_utc(ct_datetime(2001, 1, 2)),
        to_utc(ct_datetime(2002, 1, 1)),
    )
    Tpr.insert(sess, "00205", True, True)
    sess.commit()
    rows = [["0001", "00205"]]
    ctx = {}
    _import_Measurement_Requirement(sess, rows, ctx)

    mr = sess.execute(select(MeasurementRequirement)).scalar()
    assert mr.ssc == latest_ssc


def test_import_Meter_Timeswitch_Class(sess):
    rows = [
        [
            "1",
            "01/04/1996",
            "",
            "",
            "F",
            "F",
            "",
            "",
            "",
            "",
            "",
        ]
    ]
    ctx = {}
    _import_Meter_Timeswitch_Class(sess, rows, ctx)


def test_import_Meter_Timeswitch_Class_no_tpr_count(sess):
    vf = to_utc(ct_datetime(1996, 4, 1))
    MeterType.insert(sess, "C5", "A c5 meter", vf, None)
    MeterPaymentType.insert(sess, "CR", "credit", vf, None)
    rows = [
        [
            "845",
            "01/04/1996",
            "",
            "HH COP5 And Above With Comms",
            "T",
            "F",
            "C5",
            "CR",
            "Y",
            "H",
            "",
        ]
    ]
    ctx = {}
    _import_Meter_Timeswitch_Class(sess, rows, ctx)


def test_import_MTC_in_PES_Area(sess):
    vf = to_utc(ct_datetime(1996, 4, 1))
    participant_code = "HYDE"
    Participant.insert(sess, participant_code, "hyde participant")
    mtc_code = "001"
    Mtc.insert(
        sess,
        mtc_code,
        False,
        False,
        vf,
        None,
    )
    MeterType.insert(sess, "UN", "A c5 meter", vf, None)
    MeterPaymentType.insert(sess, "TS", "credit", vf, None)
    rows = [
        [
            mtc_code,
            "01/04/1996",
            "HYDE",
            "01/04/1996",
            "",
            "Unrestricted, Single rate",
            "UN",
            "TS",
            "N",
            "N",
            "1",
        ]
    ]
    ctx = {}
    _import_MTC_in_PES_Area(sess, rows, ctx)


def test_import_Time_Pattern_Regime(sess):
    rows = [["00001", "C", "N"]]
    ctx = {}
    _import_Time_Pattern_Regime(sess, rows, ctx)


def test_import_Valid_MTC_LLFC_Combination(sess):
    valid_from = to_utc(ct_datetime(1996, 4, 1))
    participant_code = "HYDE"
    participant = Participant.insert(sess, participant_code, "hyde participant")
    mtc_code = "001"
    mtc = Mtc.insert(
        sess,
        mtc_code,
        False,
        False,
        valid_from,
        None,
    )
    meter_type = MeterType.insert(sess, "UN", "A c5 meter", valid_from, None)
    meter_payment_type = MeterPaymentType.insert(sess, "TS", "credit", valid_from, None)
    MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "desc",
        True,
        True,
        meter_type,
        meter_payment_type,
        1,
        valid_from,
        None,
    )
    dno_role = MarketRole.insert(sess, "R", "dno")
    dno = participant.insert_party(sess, dno_role, "dno", valid_from, None, "10")
    insert_voltage_levels(sess)
    vl = VoltageLevel.get_by_code(sess, "HV")
    dno.insert_llfc(sess, "H85", "llfc h85", vl, False, True, valid_from, None)
    rows = [
        [
            mtc_code,
            "01/04/1996",
            participant_code,
            "01/04/1996",
            "H85",
            "01/04/2021",
            "",
        ]
    ]
    ctx = {}
    _import_Valid_MTC_LLFC_Combination(sess, rows, ctx)


def test_import_Valid_MTC_SSC_Combination(sess):
    valid_from = to_utc(ct_datetime(1996, 4, 1))
    participant_code = "HYDE"
    participant = Participant.insert(sess, participant_code, "hyde participant")
    mtc_code = "001"
    mtc = Mtc.insert(
        sess,
        mtc_code,
        False,
        False,
        valid_from,
        None,
    )
    meter_type = MeterType.insert(sess, "UN", "A c5 meter", valid_from, None)
    meter_payment_type = MeterPaymentType.insert(sess, "TS", "credit", valid_from, None)
    mtc_participant_valid_from = to_utc(ct_datetime(2005, 12, 2))
    MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "desc",
        True,
        True,
        meter_type,
        meter_payment_type,
        1,
        mtc_participant_valid_from,
        None,
    )
    ssc_code = "0721"
    Ssc.insert(sess, ssc_code, "desc", True, valid_from, None)
    rows = [
        [
            mtc_code,
            "01/04/1996",
            participant_code,
            "02/12/2005",
            ssc_code,
            "02/12/2005",
            "",
        ]
    ]
    ctx = {}
    _import_Valid_MTC_SSC_Combination(sess, rows, ctx)


def test_import_Valid_MTC_LLFC_SSC_Combination(sess):
    valid_from = to_utc(ct_datetime(1996, 4, 1))
    participant_code = "HYDE"
    participant = Participant.insert(sess, participant_code, "hyde participant")
    mtc_code = "001"
    mtc = Mtc.insert(
        sess,
        mtc_code,
        False,
        False,
        valid_from,
        None,
    )
    meter_type = MeterType.insert(sess, "UN", "A c5 meter", valid_from, None)
    meter_payment_type = MeterPaymentType.insert(sess, "TS", "credit", valid_from, None)
    mtc_participant_valid_from = to_utc(ct_datetime(2005, 12, 2))
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "desc",
        True,
        True,
        meter_type,
        meter_payment_type,
        1,
        mtc_participant_valid_from,
        None,
    )
    ssc_code = "0721"
    ssc = Ssc.insert(sess, ssc_code, "desc", True, valid_from, None)
    dno_role = MarketRole.insert(sess, "R", "dno")
    dno = participant.insert_party(sess, dno_role, "dno", valid_from, None, "10")
    insert_voltage_levels(sess)
    vl = VoltageLevel.get_by_code(sess, "HV")
    llfc_code = "183"
    dno.insert_llfc(sess, llfc_code, "llfc h85", vl, False, True, valid_from, None)
    MtcSsc.insert(sess, mtc_participant, ssc, to_utc(ct_datetime(2005, 12, 2)), None)
    rows = [
        [
            mtc_code,
            "01/04/1996",
            participant_code,
            "02/12/2005",
            ssc_code,
            "02/12/2005",
            llfc_code,
            "01/04/2006",
            "",
        ]
    ]
    ctx = {}
    _import_Valid_MTC_LLFC_SSC_Combination(sess, rows, ctx)


def test_import_Valid_MTC_LLFC_SSC_PC_Combination(sess):
    valid_from = to_utc(ct_datetime(1996, 4, 1))
    participant_code = "HYDE"
    participant = Participant.insert(sess, participant_code, "hyde participant")
    mtc_code = "001"
    mtc = Mtc.insert(
        sess,
        mtc_code,
        False,
        False,
        valid_from,
        None,
    )
    meter_type = MeterType.insert(sess, "UN", "A c5 meter", valid_from, None)
    meter_payment_type = MeterPaymentType.insert(sess, "TS", "credit", valid_from, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "desc",
        True,
        True,
        meter_type,
        meter_payment_type,
        1,
        to_utc(ct_datetime(2009, 4, 16)),
        None,
    )
    ssc_code = "0721"
    ssc = Ssc.insert(sess, ssc_code, "desc", True, valid_from, None)
    dno_role = MarketRole.insert(sess, "R", "dno")
    dno = participant.insert_party(sess, dno_role, "dno", valid_from, None, "10")
    insert_voltage_levels(sess)
    vl = VoltageLevel.get_by_code(sess, "HV")
    llfc_code = "183"
    llfc = dno.insert_llfc(
        sess, llfc_code, "llfc h85", vl, False, True, valid_from, None
    )
    pc_code = "00"
    Pc.insert(sess, pc_code, "hh", utc_datetime(2000, 1, 1), None)
    mtc_ssc = MtcSsc.insert(
        sess, mtc_participant, ssc, to_utc(ct_datetime(2009, 4, 16)), None
    )
    mtc_llfc_ssc = MtcLlfcSsc.insert(
        sess, mtc_ssc, llfc, to_utc(ct_datetime(2009, 4, 16)), None
    )
    rows = [
        [
            mtc_code,
            "01/04/1996",
            participant_code,
            "16/04/2009",
            ssc_code,
            "16/04/2009",
            llfc_code,
            "16/04/2009",
            pc_code,
            "17/03/2010",
            "20/08/2014",
            "F",
        ]
    ]
    ctx = {}
    _import_Valid_MTC_LLFC_SSC_PC_Combination(sess, rows, ctx)
    combos = mtc_llfc_ssc.mtc_llfc_ssc_pcs
    assert len(combos) == 1

    combo = combos[0]

    assert combo.valid_to == to_utc(ct_datetime(2014, 8, 20, 23, 30))


def test_import_mdd(mocker, sess):
    vf = to_utc(ct_datetime(1996, 4, 1))
    MeterType.insert(sess, "C5", "A c5 meter", vf, None)
    MeterPaymentType.insert(sess, "CR", "credit", vf, None)
    calb_participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    calb_participant.insert_party(sess, market_role_Z, "NonCore", vf, None, "")
    Contract.insert_non_core(sess, "configuration", "", {}, vf, None, {})
    sess.commit()

    log = mocker.Mock()
    set_progress = mocker.Mock()

    mocker.patch("chellow.e.mdd_importer.download", return_value=b"\n")

    path_list = (
        ("2022", "electricity", "mdd", "50", "Clock_Interval_50.csv"),
        ("2022", "electricity", "mdd", "50", "GSP_Group_50.csv"),
        ("2022", "electricity", "mdd", "50", "Line_Loss_Factor_Class_50.csv"),
        ("2022", "electricity", "mdd", "50", "Market_Participant_50.csv"),
        ("2022", "electricity", "mdd", "50", "Market_Role_50.csv"),
        ("2022", "electricity", "mdd", "50", "Market_Participant_Role_50.csv"),
        ("2022", "electricity", "mdd", "50", "Measurement_Requirement_50.csv"),
        ("2022", "electricity", "mdd", "50", "Meter_Timeswitch_Class_50.csv"),
        ("2022", "electricity", "mdd", "50", "MTC_in_PES_Area_50.csv"),
        ("2022", "electricity", "mdd", "50", "MTC_Meter_Type_50.csv"),
        ("2022", "electricity", "mdd", "50", "MTC_Payment_Type_50.csv"),
        ("2022", "electricity", "mdd", "50", "Profile_Class_50.csv"),
        (
            "2022",
            "electricity",
            "mdd",
            "50",
            "Standard_Settlement_Configuration_50.csv",
        ),
        ("2022", "electricity", "mdd", "50", "Time_Pattern_Regime_50.csv"),
        ("2022", "electricity", "mdd", "50", "Valid_MTC_LLFC_Combination_50.csv"),
        ("2022", "electricity", "mdd", "50", "Valid_MTC_SSC_Combination_50.csv"),
        ("2022", "electricity", "mdd", "50", "Valid_MTC_LLFC_SSC_Combination_50.csv"),
        (
            "2022",
            "electricity",
            "mdd",
            "50",
            "Valid_MTC_LLFC_SSC_PC_Combination_50.csv",
        ),
    )
    paths = tuple((p, "") for p in path_list)
    s = mocker.Mock()

    rate_server_import(sess, log, set_progress, s, paths)


def test_import_mdd_two_versions(mocker, sess):
    mocker.patch("chellow.e.mdd_importer.download", return_value=b"\n")

    path_list = (
        ("2022", "electricity", "mdd", "49"),
        ("2022", "electricity", "mdd", "50", "Clock_Interval_50.csv"),
        ("2022", "electricity", "mdd", "50", "GSP_Group_50.csv"),
        ("2022", "electricity", "mdd", "50", "Line_Loss_Factor_Class_50.csv"),
        ("2022", "electricity", "mdd", "50", "Market_Participant_50.csv"),
        ("2022", "electricity", "mdd", "50", "Market_Role_50.csv"),
        ("2022", "electricity", "mdd", "50", "Market_Participant_Role_50.csv"),
        ("2022", "electricity", "mdd", "50", "Meter_Timeswitch_Class_50.csv"),
        ("2022", "electricity", "mdd", "50", "Measurement_Requirement_50.csv"),
        ("2022", "electricity", "mdd", "50", "MTC_in_PES_Area_50.csv"),
        ("2022", "electricity", "mdd", "50", "MTC_Meter_Type_50.csv"),
        ("2022", "electricity", "mdd", "50", "MTC_Payment_Type_50.csv"),
        ("2022", "electricity", "mdd", "50", "Profile_Class_50.csv"),
        (
            "2022",
            "electricity",
            "mdd",
            "50",
            "Standard_Settlement_Configuration_50.csv",
        ),
        ("2022", "electricity", "mdd", "50", "Time_Pattern_Regime_50.csv"),
        ("2022", "electricity", "mdd", "50", "Valid_MTC_LLFC_Combination_50.csv"),
        ("2022", "electricity", "mdd", "50", "Valid_MTC_SSC_Combination_50.csv"),
        ("2022", "electricity", "mdd", "50", "Valid_MTC_LLFC_SSC_Combination_50.csv"),
        (
            "2022",
            "electricity",
            "mdd",
            "50",
            "Valid_MTC_LLFC_SSC_PC_Combination_50.csv",
        ),
    )
    paths = tuple((p, "") for p in path_list)

    vf = to_utc(ct_datetime(1996, 4, 1))
    MeterType.insert(sess, "C5", "A c5 meter", vf, None)
    MeterPaymentType.insert(sess, "CR", "credit", vf, None)
    calb_participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    calb_participant.insert_party(sess, market_role_Z, "NonCore", vf, None, "")
    Contract.insert_non_core(sess, "configuration", "", {}, vf, None, {})
    sess.commit()

    log = mocker.Mock()
    set_progress = mocker.Mock()
    s = mocker.Mock()
    rate_server_import(sess, log, set_progress, s, paths)



import atexit
import collections
import csv
import threading
import traceback
from datetime import datetime as Datetime
from decimal import Decimal, InvalidOperation

from dateutil.relativedelta import relativedelta

import requests

from sqlalchemy import or_
from sqlalchemy.sql.expression import null

from werkzeug.exceptions import BadRequest

from zish import dumps, loads

from chellow.models import Contract, RateScript, Session, get_non_core_contract_id
from chellow.utils import HH, hh_format, to_ct, to_utc, utc_datetime_now


ELEXON_PORTAL_SCRIPTING_KEY_KEY = "elexonportal_scripting_key"

RUNS = ["DF", "RF", "R3", "R2", "R1", "SF", "II"]


def key_format(dt):
    return dt.strftime("%d %H:%M")


def hh(data_source, run="DF"):
    gsp_group_code = data_source.gsp_group_code
    use_runs = RUNS[RUNS.index(run) :]

    try:
        cache = data_source.caches["tlms"]
    except KeyError:
        cache = data_source.caches["tlms"] = {}

    for h in data_source.hh_data:
        try:
            h["tlm"] = tlm = cache[h["start-date"]][gsp_group_code][run]
        except KeyError:
            h_start = h["start-date"]
            db_id = get_non_core_contract_id("tlms")
            rates = data_source.hh_rate(db_id, h_start)["tlms"]

            key = key_format(h_start)
            try:
                rate = rates[key]
            except KeyError:
                rate = sorted(rates.items())[-1][1]

            try:
                gsp_rate = rate[gsp_group_code]
            except KeyError:
                gsp_rate = sorted(rate.items())[-1][1]

            tlm = None
            for use_run in use_runs:
                try:
                    h["tlm"] = tlm = float(gsp_rate[use_run]["off_taking"])
                    break
                except KeyError:
                    pass

            if tlm is None:
                h["tlm"] = tlm = float(sorted(gsp_rate.items())[-1][1]["off_taking"])

            try:
                rates_cache = cache[h["start-date"]]
            except KeyError:
                rates_cache = cache[h["start-date"]] = {}

            try:
                gsp_cache = rates_cache[gsp_group_code]
            except KeyError:
                gsp_cache = rates_cache[gsp_group_code] = {}

            gsp_cache[run] = tlm

        h["nbp-kwh"] = h["gsp-kwh"] * tlm


tlm_importer = None


class TlmImporter(threading.Thread):
    def __init__(self):
        super(TlmImporter, self).__init__(name="TLM Importer")
        self.lock = threading.RLock()
        self.messages = collections.deque(maxlen=100)
        self.stopped = threading.Event()
        self.going = threading.Event()
        self.PROXY_HOST_KEY = "proxy.host"
        self.PROXY_PORT_KEY = "proxy.port"

    def stop(self):
        self.stopped.set()
        self.going.set()
        self.join()

    def go(self):
        self.going.set()

    def is_locked(self):
        if self.lock.acquire(False):
            self.lock.release()
            return False
        else:
            return True

    def log(self, message):
        self.messages.appendleft(
            utc_datetime_now().strftime("%Y-%m-%d %H:%M:%S") + " - " + message
        )

    def run(self):
        while not self.stopped.isSet():
            if self.lock.acquire(False):
                try:
                    _import_tlms(self.log)
                finally:
                    self.lock.release()

            self.going.wait(24 * 60 * 60)
            self.going.clear()


def _save_cache(sess, cache):
    for yr, yr_cache in cache.items():
        for month, (rs, rates, rts) in tuple(yr_cache.items()):
            rs.script = dumps(rates)
            sess.commit()
            del yr_cache[month]


def _import_tlms(log_func):
    cache = {}
    with Session() as sess:
        try:
            log_func("Starting to check TLMs.")
            contract = Contract.get_non_core_by_name(sess, "tlms")
            contract_props = contract.make_properties()
            if contract_props.get("enabled", False):
                config = Contract.get_non_core_by_name(sess, "configuration")
                props = config.make_properties()
                scripting_key = props.get(ELEXON_PORTAL_SCRIPTING_KEY_KEY)
                if scripting_key is None:
                    raise BadRequest(
                        f"The property {ELEXON_PORTAL_SCRIPTING_KEY_KEY} cannot be "
                        f"found in the configuration properties."
                    )

                url_str = (
                    f"{contract_props['url']}file/download/TLM_FILE?key={scripting_key}"
                )

                sess.rollback()  # Avoid long-running transaction
                r = requests.get(url_str)
                parser = csv.reader(
                    (x.decode() for x in r.iter_lines()), delimiter=",", quotechar='"'
                )
                log_func(f"Opened {url_str}.")

                next(parser, None)
                for i, values in enumerate(parser):
                    if values[3] == "":
                        for zone in GSP_GROUP_LOOKUP.keys():
                            values[3] = zone
                            _process_line(cache, sess, contract, log_func, values)
                    else:
                        _process_line(cache, sess, contract, log_func, values)

                _save_cache(sess, cache)
            else:
                log_func(
                    "The importer is disabled. Set 'enabled' to 'true' in the "
                    "properties to enable it."
                )

        except BadRequest as e:
            log_func(f"Problem: {e.description}")
            sess.rollback()
        except BaseException:
            log_func(f"Outer problem {traceback.format_exc()}")
            sess.rollback()
        finally:
            log_func("Finished checking TLM rates.")


GSP_GROUP_LOOKUP = {
    "1": "_A",
    "2": "_B",
    "3": "_C",
    "4": "_D",
    "5": "_E",
    "6": "_F",
    "7": "_G",
    "8": "_H",
    "9": "_J",
    "10": "_K",
    "11": "_L",
    "12": "_M",
    "13": "_N",
    "14": "_P",
}


def _process_line(cache, sess, contract, log_func, values):
    hh_date_ct = to_ct(Datetime.strptime(values[0], "%d/%m/%Y"))
    hh_date = to_utc(hh_date_ct)
    hh_date += relativedelta(minutes=30 * (int(values[2]) - 1))
    run = values[1]
    gsp_group_code = GSP_GROUP_LOOKUP[values[3]]
    off_taking_str = values[4]

    try:
        off_taking = Decimal(off_taking_str)
    except InvalidOperation as e:
        raise BadRequest(
            f"Problem parsing 'off-taking' field '{off_taking_str}' in the row "
            f"{values}. {e}"
        )

    delivering = Decimal(values[5])

    try:
        rs, rates, rts = cache[hh_date.year][hh_date.month]
    except KeyError:
        _save_cache(sess, cache)
        try:
            yr_cache = cache[hh_date.year]
        except KeyError:
            yr_cache = cache[hh_date.year] = {}

        rs = (
            sess.query(RateScript)
            .filter(
                RateScript.contract == contract,
                RateScript.start_date <= hh_date,
                or_(
                    RateScript.finish_date == null(), RateScript.finish_date >= hh_date
                ),
            )
            .first()
        )
        while rs is None:
            log_func("There's no rate script at " + hh_format(hh_date) + ".")
            latest_rs = (
                sess.query(RateScript)
                .filter(RateScript.contract == contract)
                .order_by(RateScript.start_date.desc())
                .first()
            )
            contract.update_rate_script(
                sess,
                latest_rs,
                latest_rs.start_date,
                latest_rs.start_date + relativedelta(months=2) - HH,
                loads(latest_rs.script),
            )
            new_rs_start = latest_rs.start_date + relativedelta(months=1)
            contract.insert_rate_script(sess, new_rs_start, {})
            sess.commit()
            log_func("Added a rate script starting at " + hh_format(new_rs_start) + ".")

            rs = (
                sess.query(RateScript)
                .filter(
                    RateScript.contract == contract,
                    RateScript.start_date <= hh_date,
                    or_(
                        RateScript.finish_date == null(),
                        RateScript.finish_date >= hh_date,
                    ),
                )
                .first()
            )

        rates = loads(rs.script)

        try:
            rts = rates["tlms"]
        except KeyError:
            rts = rates["tlms"] = {}

        yr_cache[hh_date.month] = rs, rates, rts
        sess.rollback()

    key = key_format(hh_date)
    try:
        existing = rts[key]
    except KeyError:
        existing = rts[key] = {}

    try:
        group = existing[gsp_group_code]
    except KeyError:
        group = existing[gsp_group_code] = {}

    if run not in group:
        group[run] = {"off_taking": off_taking, "delivering": delivering}

        log_func(
            "Found rate at "
            + hh_format(hh_date)
            + " for GSP Group "
            + gsp_group_code
            + " and run "
            + run
            + "."
        )


def get_importer():
    return tlm_importer


def startup():
    global tlm_importer
    if tlm_importer is not None:
        raise BadRequest("The TLM importer has already been started.")
    tlm_importer = TlmImporter()
    tlm_importer.start()


@atexit.register
def shutdown():
    if tlm_importer is not None:
        tlm_importer.stop()



import csv
import os
import sys
import threading
import traceback

from flask import g

from sqlalchemy import select
from sqlalchemy.orm import joinedload
from sqlalchemy.sql.expression import null

import chellow.dloads
from chellow.models import (
    Contract,
    Era,
    Party,
    ReportRun,
    Session,
    SiteEra,
    Source,
    Supply,
    User,
)
from chellow.views import chellow_redirect


FNAME = "supply_contacts"


def content(user_id, report_run_id):
    f = report_run = None
    try:
        with Session() as sess:
            user = User.get_by_id(sess, user_id)
            running_name, finished_name = chellow.dloads.make_names(
                f"{FNAME}.csv", user
            )
            f = open(running_name, mode="w", newline="")
            report_run = ReportRun.get_by_id(sess, report_run_id)

            _process(sess, f, report_run)
            report_run.update("finished")
            sess.commit()

    except BaseException as e:
        msg = traceback.format_exc()
        if report_run is not None:
            report_run.update("interrupted")
            report_run.insert_row(sess, "", ["problem"], {"problem": msg}, {})
            sess.commit()
        sys.stderr.write(msg)
        if f is not None:
            f.write(msg)
        raise e
    finally:
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def _process(sess, f, report_run):
    writer = csv.writer(f, lineterminator="\n")

    titles = (
        "site",
        "dno_contact",
        "import_mpan_core",
        "import_supplier_contact",
        "problem",
    )
    writer.writerow(titles)

    configuration = Contract.get_non_core_by_name(sess, "configuration")
    props = configuration.make_properties()
    dno_contacts = props["dno_contacts"]
    dno_contact_message = props.get("dno_contact_message")
    run_data = report_run.data
    run_data["dno_contact_message"] = dno_contact_message
    report_run.update_data(run_data)
    sess.commit()

    for era in (
        sess.execute(
            select(Era)
            .join(Supply)
            .join(Supply.dno)
            .join(Source)
            .where(
                Era.finish_date == null(),
                Party.dno_code.notin_(("88", "99")),
                Era.finish_date == null(),
                Source.code != "3rd-party",
            )
            .order_by(Era.supply_id)
            .options(
                joinedload(Era.site_eras).joinedload(SiteEra.site),
                joinedload(Era.supply).joinedload(Supply.dno),
                joinedload(Era.imp_supplier_contract),
                joinedload(Era.mop_contract),
            )
        )
        .scalars()
        .unique()
    ):
        for site_era in era.site_eras:
            if site_era.is_physical:
                site = site_era.site

        dno = era.supply.dno

        imp_supplier_contract = era.imp_supplier_contract
        if imp_supplier_contract is None:
            imp_supplier_name = None
        else:
            imp_supplier_name = era.imp_supplier_contract.party.name

        values = {
            "site_code": site.code,
            "site_name": site.name,
            "dno_name": dno.name,
            "dno_contact": dno_contacts.get(dno.dno_code),
            "mop_name": era.mop_contract.party.name,
            "import_mpan_core": era.imp_mpan_core,
            "import_supplier_name": imp_supplier_name,
        }
        report_run.insert_row(sess, "", titles, values, {})
        sess.commit()


def do_get(sess):
    report_run = ReportRun.insert(
        sess,
        FNAME,
        g.user,
        FNAME,
        {},
    )
    sess.commit()
    threading.Thread(target=content, args=(g.user.id, report_run.id)).start()
    return chellow_redirect(f"/report_runs/{report_run.id}", 303)



import csv
import os
import threading
import traceback

from flask import g

from sqlalchemy import or_
from sqlalchemy.sql.expression import null, true

from werkzeug.exceptions import BadRequest

import chellow.dloads
from chellow.e.computer import SupplySource, forecast_date
from chellow.models import (
    Era,
    MeasurementRequirement,
    Session,
    Site,
    SiteEra,
    Ssc,
    Supply,
    Tpr,
)
from chellow.utils import csv_make_val, hh_format, hh_max, hh_min, req_date, req_int
from chellow.views import chellow_redirect


def content(supply_id, file_name, start_date, finish_date, user):
    caches = {}
    try:
        with Session() as sess:
            running_name, finished_name = chellow.dloads.make_names(
                f"supply_virtual_bills_{supply_id}.csv", user
            )
            f = open(running_name, mode="w", newline="")
            writer = csv.writer(f, lineterminator="\n")

            supply = Supply.get_by_id(sess, supply_id)

            forecast_dt = forecast_date()

            prev_titles = None

            for era in (
                sess.query(Era)
                .filter(
                    Era.supply == supply,
                    Era.start_date < finish_date,
                    or_(Era.finish_date == null(), Era.finish_date > start_date),
                )
                .order_by(Era.start_date)
            ):
                chunk_start = hh_max(era.start_date, start_date)
                chunk_finish = hh_min(era.finish_date, finish_date)
                site = (
                    sess.query(Site)
                    .join(SiteEra)
                    .filter(SiteEra.era == era, SiteEra.is_physical == true())
                    .one()
                )

                ds = SupplySource(
                    sess,
                    chunk_start,
                    chunk_finish,
                    forecast_dt,
                    era,
                    era.imp_supplier_contract is not None,
                    caches,
                )

                titles = [
                    "Imp MPAN Core",
                    "Exp MPAN Core",
                    "Site Code",
                    "Site Name",
                    "Account",
                    "From",
                    "To",
                    "",
                ]

                output_line = [
                    era.imp_mpan_core,
                    era.exp_mpan_core,
                    site.code,
                    site.name,
                    ds.supplier_account,
                    hh_format(ds.start_date),
                    hh_format(ds.finish_date),
                    "",
                ]

                mop_titles = ds.contract_func(era.mop_contract, "virtual_bill_titles")()
                titles.extend(["mop-" + t for t in mop_titles])

                ds.contract_func(era.mop_contract, "virtual_bill")(ds)
                bill = ds.mop_bill
                for title in mop_titles:
                    if title in bill:
                        output_line.append(bill[title])
                        del bill[title]
                    else:
                        output_line.append("")

                for k in sorted(bill.keys()):
                    output_line.extend([k, bill[k]])

                output_line.append("")
                dc_titles = ds.contract_func(era.dc_contract, "virtual_bill_titles")()
                titles.append("")
                titles.extend(["dc-" + t for t in dc_titles])

                ds.contract_func(era.dc_contract, "virtual_bill")(ds)
                bill = ds.dc_bill
                for title in dc_titles:
                    output_line.append(bill.get(title, ""))
                    if title in bill:
                        del bill[title]
                for k in sorted(bill.keys()):
                    output_line.extend([k, bill[k]])

                tpr_query = (
                    sess.query(Tpr)
                    .join(MeasurementRequirement)
                    .join(Ssc)
                    .join(Era)
                    .filter(
                        Era.start_date <= chunk_finish,
                        or_(Era.finish_date == null(), Era.finish_date >= chunk_start),
                    )
                    .order_by(Tpr.code)
                    .distinct()
                )

                if era.imp_supplier_contract is not None:
                    output_line.append("")
                    supplier_titles = ds.contract_func(
                        era.imp_supplier_contract, "virtual_bill_titles"
                    )()
                    for tpr in tpr_query.filter(Era.imp_supplier_contract != null()):
                        for suffix in ("-kwh", "-rate", "-gbp"):
                            supplier_titles.append(tpr.code + suffix)
                    titles.append("")
                    titles.extend(["imp-supplier-" + t for t in supplier_titles])

                    ds.contract_func(era.imp_supplier_contract, "virtual_bill")(ds)
                    bill = ds.supplier_bill

                    for title in supplier_titles:
                        if title in bill:
                            output_line.append(bill[title])
                            del bill[title]
                        else:
                            output_line.append("")

                    for k in sorted(bill.keys()):
                        output_line.extend([k, bill[k]])

                if era.exp_supplier_contract is not None:
                    ds = SupplySource(
                        sess, chunk_start, chunk_finish, forecast_dt, era, False, caches
                    )

                    output_line.append("")
                    supplier_titles = ds.contract_func(
                        era.exp_supplier_contract, "virtual_bill_titles"
                    )()
                    for tpr in tpr_query.filter(Era.exp_supplier_contract != null()):
                        for suffix in ("-kwh", "-rate", "-gbp"):
                            supplier_titles.append(tpr.code + suffix)
                    titles.append("")
                    titles.extend(["exp-supplier-" + t for t in supplier_titles])

                    ds.contract_func(era.exp_supplier_contract, "virtual_bill")(ds)
                    bill = ds.supplier_bill
                    for title in supplier_titles:
                        output_line.append(bill.get(title, ""))
                        if title in bill:
                            del bill[title]

                    for k in sorted(bill.keys()):
                        output_line.extend([k, bill[k]])

                if titles != prev_titles:
                    prev_titles = titles
                    writer.writerow([str(v) for v in titles])
                for i, val in enumerate(output_line):
                    output_line[i] = csv_make_val(val)
                writer.writerow(output_line)
    except BadRequest as e:
        writer.writerow(["Problem: " + e.description])
    except BaseException:
        writer.writerow([traceback.format_exc()])
    finally:
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    supply_id = req_int("supply_id")
    file_name = "supply_virtual_bills_" + str(supply_id) + ".csv"
    start_date = req_date("start")
    finish_date = req_date("finish")
    args = (supply_id, file_name, start_date, finish_date, g.user)

    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



from flask import g, render_template

from chellow.e.bill_importer import BillImport, find_parser_names


def test_status(mocker):
    batch = mocker.Mock()
    bill_import = BillImport(batch)
    bill_import.isAlive = mocker.Mock(return_value=True)
    bill_import.status()


def test_supplier_bill_import_html(mocker, app):
    with app.app_context():
        with app.test_request_context():
            g.user = None
            g.config = {}
            batch = mocker.Mock()
            failed_bills = [{"error": "MPAN not found"}]
            render_template(
                "/e/supplier_bill_import.html", batch=batch, failed_bills=failed_bills
            )


def test_find_parser_names():
    assert len(find_parser_names()) > 0
